/////////////////////////////////////////////////////////////////////////////
//
// CRT TEXT DISPLAY GENERATOR v1.0  05/10/2001
// --------------------------
//
//   Purpose : To demonstrate a possible FPGA implementation for an 
//             alphanumeric text-display controller.  This module will run
//             on the Xess XS40 v1.4 FPGA-board, without modification.
//
//   Design  : Module is coded in Verilog HDL (synthesizable RTL)
//     Entry    
//
//   Specifications & "features": 
//
//     1) display-resolution = 640x480 60Hz (31.5KHz VGA) : 25.00MHz pixclk
//        alphanumeric text display 80x30 characters
//        fixed, monochrome palette : foreground=white, background=black
//
//     2) font-resolution = 8 x YY character cell (YY=programable, 1..31)
//        font-depth = 256 addressable character bitmaps
//        (default 'VGA' fontmap = 8x16, 256 ANSI character set)
//
//     3) simple SRAM access unit - ASYNCHRONOUS sram, 8-bit wide data bus
//        (for synchronous SRAM, the memory-input bus register must be 
//         removed.)  
//        
//     4) synthesizable Verilog HDL (RTL)
//        "ASIC-style" coding, single master clock for all flip-flops.
//        Conventional single-edge triggered flipflops, no latches.
//        (This not the most efficient style for an FPGA-architecture.)
//
//     5) Pipelined architecture for faster FPGA execution 
//        combinational logic clouds distributed over several clock periods,
//        instead of compressed into 1 clock.
//
//     6) Tested to work on an Xess XS40-010XL, 25MHz master clock
//        
//   Missing logic/features :
//
//    1) no host bus interface for programming various CRTC timing generators
//       Address-map, bus-interface logic (tristate?), etc. eat resources.
//
//    2) no pixel-addressable (graphics) display mode
//       alphanumeric mode only (text)
//
//    3) no 'attribute data' controller
//       In CGA/EGA/VGA displays, each text-display character comprises
//       both the character-code byte and an attribute-byte.  The
//       attribute byte encodes a background color, a foreground color, 
//       and a fontmap-select bit.
//
/////////////////////////////////////////////////////////////////////////////
// -----
//
//   liaor@iname.com - http://members.tripod.com/~liaor  (05/01/2001)
//
//   This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., http://www.gnu.org
//
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
//
//   Top-level overview:
//   The text-controller has 4 major components :
//
//    1) horizontal/vertical timing (pulse) generator
//       Generate MAJOR timing signals
//         chartim_h_inc, chartim_v_inc   (sub-character pulse, pixel)
//         charctr_h_inc, charctr_v_inc   (text-grid pulses)
//         linectr_v_inc                  (scanline pulse)
//
//       Generate MAJOR output signals
//         hsync, vsync - VGA monitor control lines 
//         scanena - internal masking control ('active display area')
//
//    2) Text-screen access unit
//       Calculate framebuffer address for current scan-location
//       Retrieve screen (character-code) data from framebuffer
//       assumes framebuffer memory is 1T-asynchronous SRAM
//
//    3) font-bitmap access unit
//       Calculate bitmap address for current character-code & position
//       Retrieve font (mono bitmap) data from framebuffer
//       assumes framebuffer memory is 1T-asynchronous SRAM
//
//    4) Output pixel driver (display assembly)
//       Convert bitmap pixel-data into R,G,B intensity values
//       Synchronize R,G,B with hsync, vsync pulses
//       Send to module outputs (to VGA monitor)
//
//    *This module assumes the framebuffer is (ordinary) asynchronous 
//    SRAM, 1 byte-wide.  The XS40 has a single asynchronous SRAM chip,
//    32KB x 8, 20ns or faster.
//
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
//
//  Author's Developmental notes
//  ----------------------------
//
//    I designed this module to demonstrate a simple VGA character 
// controller.  First I read through an old Cirrus Logic VGA databook, to
// get a general idea how the VGA register map is setup.  Then, I coded a
// simple pixel graphics display, to gain understanding of the CRTC counters.
// Finally, I jumped into the character-text controller portion, which
// include the framebuffer-lookup and font-lookup units. 
//    Ideally, all aspects of the controller should be programmable, but 
// the final design must be practical.  To keep the design a reasonable
// size, I dropped some features in favor of economy.  At the same time, I
// wanted to partially optimize for speed.  There were two choices : 
// optimize for a particular device (device/tool-specific optimization),
// or optimize at the RTL-level.  I chose the latter approach.  The former
// would outperform the latter approach, but only for the target device.
// More importantly, the latter approach is more human-readable, and
// potentially accelerate a broader range of FPGA-architectures (but to a 
// lesser degree.)  
//    The final text-controller design (barely) fits in an XS40-05XL.
// For testing, I downloaded the design to an Xess XS40-010XL+ v1.4 board.  
// Of course, any XS40 v1.4 board will work.  (Earlier revisions cannot
// generate the 25MHz clock frequency, so they can't be used with this
// design.)  The text-controller occupies ~33% of an XS4010XL.
//
//  Technical notes
//  ---------------
//
//    The CRTC-controller has a fixed character-width of 8 pixclks.  
// To support widths larger than 8-pixclks, the controller must process 
// font-character-maps wider than one-byte.  This is a lot of extra logic.
// VGA text controllers use a cheap hack - they use an 8-wide fontmap
// and insert a blank pixel to the right.  The text-grid is spaced at
// 9x16 pixels per character, but the true font-map resolution remains 8x16.
// On the plus side, this controller already supports narower fonts 
// (<8-pixclks),  The font-fetch unit always uses the familiar byte-wide 
// bitmap (8-pixels), but the rightmost bits are unused.  For example, a 
// 4-pixclk wide font still requires 1-byte storage per scanline per 
// character-map -- only half of the bits are scanned/used.  Smaller widths
// affect the CRTC timing-values, and may require a wider horizontal counter.
//
//    The screen-access and font-access units use an aggressive timing 
// pattern with the XS40's asynchronous SRAM.  At 1pixclk per read, the
// SRAMs perform well at the default pixclk=25MHz.  At pixclk=50MHz, the
// access-units must be slowed down to 2T access cycle.  Also, the module's
// total memory bandwidth usage can be measured by examining the output 
// 'o_sram_reqn.'  o_sram_reqn only asserts (active-low) when the controller
// needs data.  Since the font-access and screen-access units each make
// 1 request per 8pixclks, the module consumes 25% of the available BW.
//
//    The default CRTC parameters approximate standard VGA timing
// (640x480 60Hz.)  Any modern digital multisync monitor can deal with
// the timing deviation, but older VGA monitors may not! 
//
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
//  Verilog Coding Style Conventions
//  --------------------------------
//    In order to understand this design, you'll need a basic familiarity 
//  with Verilog RTL coding.  For example, you know how to identify 
//  flipflops ( @ posedge clk ...), you know how to identify behavioral-
//  coded combinational logic ( always @ <sensitivity list> ), etc.
//    I have gone to great lengths to maintain a semi-consistent naming
//  rule for signal-nodes within the module.  Well, it's more like a loose 
//  guideline than a strict rule.  Regardless, consistent naming cuts down 
//  on the number of 'random nets' in the entire design, helping others
//  to read/learn the design-architecture.
//    Here are the guidelines I tried to follow :
//
//    1) Prefixs with 'pX_' : <X> = the pipeline stage #
//       Pipeline stages always flow from left to right (lower to higher #s)
//                            _                       _             _
//         |"P1" stage | --> | | -> |"P2" stage | -> | | -> ... -> | |
//         |comb. logic|     >_|    |comb. logic|    >_|           >_|
//                           reg                     reg           reg
//
//       If the same signal-name appears in more than one stage (like above),
//       then the two signals are identical, but delayed with respect to
//       each other.  
//
//         eg.  p1_chartim_h -> p2_chartim_h -> p3_chartim_h
//
//    2) Other Prefixs 'i_'/'o_' : 
//
//        i_ : module input port
//        o_ : module output port 
//        rstn: global device reset (active low)
//
//    3) Postfix 'n' : 'negative polarity' (active low)
//        Negative-polarity simply means '0'=true, '1'=false
//        eg.  hsync_n = '0: in hsync', '1: in active display, not hsync'
//        
/////////////////////////////////////////////////////////////////////////////

// EGA-textmode is 640x350 ??Hz, 8x14-character (80x25 grid.)
// VGA-textmode is 720x400 70Hz, 9x16-character (80x25 grid.)
//     (The 9x16 grid uses the 8x16 fontmap, with 1-pixel blank space inserted
//      between adjacent characters.)  Some apps use 640x480 60Hz, 80x30 grid.



// known bugs : 

//  CR_XX register map lifted from IBM VGA standard

/*            CRTC REGISTER DEFINITIONS
              -------------------------
              
              |<----- CR0 ( H_TOTAL ) -------------------------------------->|
              |                                                              |
              |<----- CR3 ( H_BLANKING_END ) --------------------------->|   |
              |                                                          |   |
              |<----- CR5 ( H_RETRACE_END ) ------------------------>|   |   |
              |                                                      |   |   |
              |<----- CR4 ( H_RETRACE_START ) ------------------>|   |   |   |
              |                                                  |   |   |   |
              |<----- CR2 ( H_BLANKING_START ) ------------->|   |   |   |   |
              |                                              |   |   |   |   |
              |<----- CR1 ( H_DISPLAY_END ) ------------>|   |   |   |   |   |
------------- +------------------------------------------+---+---+   +---+---+
 ^ ^ ^ ^ ^ ^  | +------->------->------->------>------>-+|   |   |   |   |   |
 | | | | | |  |                                        / |       |   |       |
 | | | | | |  |  ...................................../  |   |   |   |   |   |
 | | | | | |  | /            ACTIVE DISPLAY              |       |   |       |
 | | | | | |  | +------->------->------->------>------>-+| G | H | I | J | K |
 | | | | | |  |                  SCREEN                / |       |   |       |
 | | | | | |  |  ...................................../  |   |   |   |   |   |
 A B C D E F  | /                                        |       |   |       |
 | | | | | |  | +------->------->------->------>------>-+|   |   |   |   |   |
 | | | | | V  |                                          |       |   |       |
 | | | | |--- +------------------------------------------+   |   |   |   |   |
 | | | | V    |          BOTTOM      BORDER                      |   |       |
 | | | |---   +- - - - - - - - - - - - - - - - - - - - - - - +   |   |   |   |
 | | | V      |          DISPLAY BLANKED                         |   |       |
 | | |---     +--------------------------------------------------+   +---+---+
 | | V        |          VERTICAL RETRACE (SYNC)                         |     
 | |---       +--------------------------------------------------+   +---+---+
 | V          |          DISPLAY BLANKED                         |   |   |   |
 |---         +- - - - - - - - - - - - - - - - - - - - - - - - - + - + - +   |
 V            |          TOP BORDER                              |   |       |
---           +--------------------------------------------------+---+-------+
 
 (the raster-trace pattern in the above diagram is misleading.  The
  electron-beam actually extends all the way to horizontal retrace, 'I')
 
 A= CR6          (V TOTAL)
 B= CR16         (V BLANKING END)
 C= CR11         (V RETRACE END)
 D= CR10         (V RETRACE START)
 E= CR15         (V BLANKING START)
 F= CR12         (V DISPLAY END)
 G= RIGHT BORDER
 H= DISPLAY BLANKED
 I= HORIZONTAL RETRACE (SYNC)
 J= DISPLAY BLANKED
 K= LET BORDER
 
 CR13 = CRTC row stride (pitch)  (length of 1 scanline in display memory)
 CR0C:CR0D = high:low screen start address
*/


// PIXCLK = 25.000MHz  (IBM VGA PIXCLK rate = 25.18MHz)
//
// Active-display area is 640H x 480V pixels (80x30 character cells)
// Font-char = 8x16 pixels
// scan rate H = 31.5KHz, V = 60.0Hz (approximately)
//
// | For true VGA-text mode (pixclk=28.325MHz), the active-display area is |
// | 720Hx400V (80x25 character cells), font-char = 9x16.                  |
// | scan rate H = 31.5KHz, V= 70.0Hz (approximately)                      |
//
// Because the XS40 can only generate a 25MHz pixclk (and not 28MHz),
// this module programs a 640x480 60Hz pixel display.  With the default
// font-size of 8x16 pixels, the text-screen dimensions are 80x30 chars.
//


// *****************************************************************
// *********                                               *********
// ********* NOT GUARANTEED TO BE "Standard VGA frequency" *********
// *********                                               *********
// *****************************************************************

//
// Values are in units of the character_clock (not pixel_clock!)
// 1 h_character_clock = 8 pixel_clocks
// 1 scanline period = HTOTAL character_clcoks
// 1 v_character_clock = 16 scanlines

// [ real_value = 'CRXX'+1 ]  [ 'CRXX' = real_value - 1 ]

//`define DFV_REG_CR06   (525-1)  // VTOTAL, for pixclk=25.18MHz, 525 scanlines
`define DFV_REG_CR06   (526-1)  // VTOTAL, for pixclk=25.0MHz, 526 scanlines
`define DFV_REG_CR16   (521-1)  // V_BLANK_END
`define DFV_REG_CR11   (494-1)  // V_RETRACE_END
`define DFV_REG_CR10   (492-1)  // V_RETRACE_START
`define DFV_REG_CR15   (485-1)  // V_BLANK_START
`define DFV_REG_CR12   (480-1)  // V_DISPLAY_END (for VGA cr09=16-1)
//`define DFV_REG_CR12   (476-1)  // V_DISPLAY_END (for EGA cr09=14-1)

//`define DFV_REG_CR00   ((800>>3)-1) // H_TOTAL (for pixclk=25.18MHz)
`define DFV_REG_CR00   ((792>>3)-1) // H_TOTAL (for pixclk=25.00MHz)
`define DFV_REG_CR03   ((784>>3)-1) // H_BLANK_END
`define DFV_REG_CR05   ((768>>3)-1) // H_RETRACE_END
`define DFV_REG_CR04   ((656>>3)-1) // H_RETRACE_START
`define DFV_REG_CR02   ((648>>3)-1) // H_BLANK_START
//`define DFV_REG_CR01   ((640>>3)-1) // drops rightmost char of each row?!?
`define DFV_REG_CR01   ((648>>3)-1) // H_DISPLAY_END, 81 chars

`define DFV_REG_CR0C  (8'h10)   // framebuffer start address (HIGH)
`define DFV_REG_CR0D  (8'h00)   // framebuffer start address (LOW)
`define DFV_REG_CR13  (10'd080)  // +80byte offset per scanline

//`define DFV_REG_CR09  (14-1)   // EGA character cell height = 14
`define DFV_REG_CR09  (16-1)   // VGA character cell height = 16

`define DFV_REG_SR03 3'h0  // font-base address = 0x0000 (8KB granularity)
//`define DFV_REG_SR03 3'h1  // font-base address = 0x2000

// BIF_xxx = bus interface for register XXX, not used
//`define BIF_REG_CR6  cr06
//`define BIF_REG_CR16 cr16
//`define BIF_REG_CR11 cr11
//`define BIF_REG_CR10 cr10
//`define BIF_REG_CR15 cr15


`include "mult8x5u.v" // 8x5 unsigned multiplier, address calculation

module vgacrt( 
      //inputs
	pixclk,	rstn,
	i_sramd,
	
      //outputs
	o_vga_vsync_n, o_vga_hsync_n, o_vga_pixeldata,
	o_srama, o_sram_reqn
	);

parameter HCW = 7;   // horionztal counter width (7-bit = 127 characters)
parameter VCW = 10;  // vertical counter width (10-bit = 1023 scanlines)
parameter PIXELS_PER_CHAR = 7;  // 7+1=8 pixels per "character"

input         pixclk;   // global master clock ('pixel clock', 25.00MHz)
input         rstn;     // global async reset (active low)
input [7:0]   i_sramd;  // from SRAM, read data

// display control signals
output        o_vga_vsync_n; // to CRT, vertical sync (active low)
output        o_vga_hsync_n; // to CRT, horizontal sync (active low)

//output vo_scanena; // enable character scan (active display area)
output [5:0]  o_vga_pixeldata; // to CRT, [R,G,B] pixel intensity 
output [15:0] o_srama;       // to SRAM, address
output        o_sram_reqn;   // to SRAM, output-request/enable (active low)

////////////////////////////////////////////////////////////////////////////
//
// CRTC static registers, display-timing parameters 
//
/*
  reg [ (HCW-1):0 ] cr00;  // H_TOTAL
  reg [ (HCW-1):0 ] cr01;  // H_DISPLAY_END
  reg [ (HCW-1):0 ] cr02;  // H_BLANKING_START
  reg [ (HCW-1):0 ] cr03;  // H_BLANKING_END
  reg [ (HCW-1):0 ] cr04;  // H_RETRACE_START
  reg [ (HCW-1):0 ] cr05;  // H_RETRACE_END
  
  reg [ (VCW-1):0 ] cr06;  // V_TOTAL
  reg [ 4:0 ]       cr09;  // character cell height
  reg [ (VCW-1):0 ] cr10;  // V_RETRACE_START
  reg [ (VCW-1):0 ] cr11;  // V_RETRACE_END
  reg [ (VCW-1):0 ] cr12;  // V_DISPLAY_END
  reg [ (VCW-1):0 ] cr15;  // V_BLANKING_START
  reg [ (VCW-1):0 ] cr16;  // V_BLANKING_END
  reg [ 9:0 ]       cr13; // framebuffer 'pitch' (row stride)
  
  reg [ 2:0 ]       sr03; // character map select (chose font baseaddress)
*/

  reg [ 4:0 ]       cr09;  // character cell height
      
  reg [ 7:0 ]       cr0c; // framebuffer start address : HIGH byte
//  reg [ 7:0 ]       cr0d; // framebuffer start address : LOW byte

////////////////////////////////////////////////////////////////////////////
//
// CRTC state registers, current video-beam position 
//
  
  reg  [3:0 ]       p1_chartim_h;// horizontal character cell counter
  reg  [3:0 ]       p2_chartim_h;//
  reg  [(HCW-1):0 ] p2_charctr_h;// horizontal character counter (timer)
  reg  [(HCW-1):0 ] p3_charctr_h;//
  
  wire [(VCW-1):0 ] p1_linectr_v;// vertical scanline counter (timer)
  reg  [(VCW-1):0 ] p2_linectr_v;//
  reg  [(VCW-1):0 ] p3_linectr_v;//
  reg  [ 4:0 ]      p3_chartim_v;// vertical character cell timer (divider)
  reg  [ 5:0 ]      p3_charctr_v;// vertical character cell counter
  
  wire [ 9:0 ]      p2_pixctr_h; // hpix counter, (char_size_h must be 8!)
  reg  [ 9:0 ]      p3_pixctr_h; //
  
  reg p3_pixena_h;  // (status) charctr_h is in "active display area"
  reg p4_pixena_h;  
  reg p4_pixena_v;  // (status) linectr_v is in "active display area" 

////////////////////////////////////////////////////////////////////////////
//
// ctl_xxx - control lines, these wires/regs direct the CRT scanning process
// 
  wire p2_charctr_h_inc;   // increment horizontal CHAR counter (not pixctr_h)
  wire p2_charctr_h_reset; // reset horizontal CHARACTER counter (not pixctr_h)
  reg  p3_charctr_h_inc;   // 
  reg  p3_charctr_h_reset; //  

  wire p3_linectr_v_inc;   // increment (vertical) scanline counter (not CHAR_V)
  wire p3_linectr_v_reset; // reset (vertical) scanline counter (not CHAR_V)
  reg  p4_charctr_h_inc;   //  
  reg  p4_charctr_h_reset; //  
  reg  p4_linectr_v_inc; 
  reg  p4_linectr_v_reset; 

  wire p1_chartim_h_reset;// reset character cell timer (horizontal)
  wire p1_chartim_h_inc;  // increment character cell timer (horizontal)
  reg  p2_chartim_h_reset;// reset character cell timer (horizontal)
  reg  p2_chartim_h_inc;  // increment character cell timer (horizontal)

  wire p3_chartim_v_reset;// reset character cell timer (vertical)
  wire p3_chartim_v_inc;  // increment character cell timer (vertical)
  wire p3_charctr_v_reset;// reset character cell counter (vertical)
  wire p3_charctr_v_inc;  // increment character cell counter (vertical)
  reg  p4_chartim_v_reset;
  reg  p4_chartim_v_inc;

  wire p3_pixena_h_set;   // enable horizontal scanout
  wire p3_pixena_h_reset; // disable horizontal scanout

  wire p4_pixena_v_set;   // enable vertical scanout
  wire p4_pixena_v_reset; // disable vertical scanout

  reg  p3_hpulse;  // precursor for vo_hsync
  wire p3_hpulse_set;
  wire p3_hpulse_reset;

  reg  p4_hpulse;  // precursor for vo_hsync
  reg  p4_vpulse;  // precursor for vo_vsync
  wire p4_vpulse_set;
  wire p4_vpulse_reset;
  wire p4_scanena; // precursor for vo_scanena
  
  // derived logic equations (comparisons)
  wire p1_h_eq_ht_m1;  // (charctr_h == "cr00 - 1")
  
  wire p2_h_eq_hbe;    // (charctr_h == cr03)   horizontal_blank_end
  wire p2_h_eq_hre;    // (charctr_h == cr05)   horizontal_retrace_end
  wire p2_h_eq_hrs;    // (charctr_h == cr04)   horizontal_retrace_start
  wire p2_h_eq_hbs;    // (charctr_h == cr02)   horizontal_blank_start
  wire p2_h_eq_hde;    // (charctr_h == cr01)   horizontal_display_end
  reg  p2_h_eq_ht_m1;  // (charctr_h == "cr00 - 1")  horizontal_total
  wire p2_v_eq_vt_m1;  // (linectr_v == "cr06 - 1")  vertical_total
  
  wire p3_v_eq_vbe;    // (linectr_v == cr16)   vertical_blank_end
  wire p3_v_eq_vre;    // (linectr_v == cr11)   vertical_retrace_end
  wire p3_v_eq_vrs;    // (linectr_v == cr10)   vertical_retrace_start
  wire p3_v_eq_vbs;    // (linectr_v == cr15)   vertical_blank_start
  wire p3_v_eq_vde;    // (linectr_v == cr12)   vertical_display_end
  
  reg  p3_h_eq_hbe;    // (charctr_h == cr03)
  reg  p3_h_eq_hre;    // (charctr_h == cr05)
  reg  p3_h_eq_hrs;    // (charctr_h == cr04)
  reg  p3_h_eq_hbs;    // (charctr_h == cr02)
  reg  p3_h_eq_hde;    // (charctr_h == cr01)
  reg  p3_v_eq_vt_m1;  // (linectr_v == "cr06 - 1") 
  
  reg  p4_v_eq_vbe;    // (linectr_v == cr16)
  reg  p4_v_eq_vre;    // (linectr_v == cr11)
  reg  p4_v_eq_vrs;    // (linectr_v == cr10)
  reg  p4_v_eq_vbs;    // (linectr_v == cr15)
  reg  p4_v_eq_vde;    // (linectr_v == cr12)
  
////////////////////////////////////////////////////////////////////////////
//
// framebuffer address logic 
//
  reg  [15:0] fbs_fbbase; // framebuffer/scanline base address
  wire [15:0] fbs_fbbase_next; // next state version 
  reg  [15:0] fbs_fbbase_next_1e; // next state version 
  reg  [15:0] fbs_fbaddr; // framebuffer memory address (base + char_offset)

  wire fbs_fbbase_inc;       // control, add scanline_pitch to fbs_fbbase
  wire fbs_fbbase_reset;     // control, reset fbaddr to top of screen

  reg  fbs_fbaddr_charinc;   // control, increment fbs_fbaddr +1 
  reg  fbs_fbaddr_reload;    // control, reload fbaddr from fbbase

  wire fbs_fbaddr_charinc_1e;// 1T early version
  wire fbs_fbaddr_reload_1e; // 1T early version

  reg  fbs_scanena;          // gating control, enable RGB pixel scanonut

  // font-map address multiplier 
  wire [7:0] font_inx;
  wire [7:0] font_inso;
  wire [4:0] font_iny; 
  wire font_start, font_busy, font_done;
  wire [12:0] font_outz;
  wire [15:0] font_addr;

  // SRAM-address output mux
  wire        o_srama_sel;
  wire [15:0] o_srama_next;
  reg  [15:0] o_srama;
  reg         o_sram_reqn;

  wire f0_stb;
  reg  f1_stb, f2_stb, f3_stb;
  reg  [7:0] f2_text_data, f3_font_data;
  reg  [12:0] f1_font_addr;
  reg [3:0] ds_vsync, ds_hsync; // deskew registers
  reg [3:1] ds_scanena;         // deskew registers

  // module output regs
  reg  [5:0] vo_pixeldata; // pixel value (4level intensity)
  wire [1:0] vo_pixeldataR; // RED value (4level intensity)
  wire [1:0] vo_pixeldataG; // GREEN value (4level intensity)
  wire [1:0] vo_pixeldataB; // BLUE value (4level intensity)


////////////////////////////////////////////////////////////////////////////
//
// CRTC "static" register bank
//

// The CRTC registers "crXX" store display parameters.  
// The display-sequencer uses the registers as read-only, to reload timers.
// The host-interface should have R/W access to these the CRTC 
// register-bank.

// CR_HORIZONTAL registers
/*
always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    begin
      cr00 <= `DFV_REG_CR00;
      cr01 <= `DFV_REG_CR01;
      cr02 <= `DFV_REG_CR02;
      cr03 <= `DFV_REG_CR03;
      cr04 <= `DFV_REG_CR04;
      cr05 <= `DFV_REG_CR05;
    end
  else
    begin
      cr00 <= `DFV_REG_CR00;
      cr01 <= `DFV_REG_CR01;
      cr02 <= `DFV_REG_CR02;
      cr03 <= `DFV_REG_CR03;
      cr04 <= `DFV_REG_CR04;
      cr05 <= `DFV_REG_CR05;
    end

// CR_VERTICAL registers
always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    begin
      cr06 <= `DFV_REG_CR06;
      cr10 <= `DFV_REG_CR10;
      cr11 <= `DFV_REG_CR11;
      cr12 <= `DFV_REG_CR12;
      cr15 <= `DFV_REG_CR15;
      cr16 <= `DFV_REG_CR16;
      cr09 <= `DFV_REG_CR09;
    end
  else
    begin
      cr06 <= `DFV_REG_CR06;
      cr10 <= `DFV_REG_CR10;
      cr11 <= `DFV_REG_CR11;
      cr12 <= `DFV_REG_CR12;
      cr15 <= `DFV_REG_CR15;
      cr16 <= `DFV_REG_CR16;
      cr09 <= `DFV_REG_CR09;
    end
*/
always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    cr09 <= (`DFV_REG_CR09);
  else
    cr09 <= (`DFV_REG_CR09);

//parameter cr09 = (`DFV_REG_CR09);

parameter cr00 = (`DFV_REG_CR00);
parameter cr01 = (`DFV_REG_CR01);
parameter cr02 = (`DFV_REG_CR02);
parameter cr03 = (`DFV_REG_CR03);
parameter cr04 = (`DFV_REG_CR04);
parameter cr05 = (`DFV_REG_CR05);
parameter cr06 = (`DFV_REG_CR06);

parameter cr10 = (`DFV_REG_CR10);
parameter cr11 = (`DFV_REG_CR11);
parameter cr12 = (`DFV_REG_CR12);
parameter cr15 = (`DFV_REG_CR15);
parameter cr16 = (`DFV_REG_CR16);

parameter cr0d = (`DFV_REG_CR0D);

parameter sr03 = (`DFV_REG_SR03); // font-location (8KB granularity) 

// Framebuffer address, offset registers
always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    begin
      cr0c <= `DFV_REG_CR0C;
//      cr0d <= `DFV_REG_CR0D;
//      cr13 <= `DFV_REG_CR13;
    end
  else
    begin
      cr0c <= `DFV_REG_CR0C;
//      cr0d <= `DFV_REG_CR0D;
//      cr13 <= `DFV_REG_CR13;
    end

parameter cr13 = (`DFV_REG_CR13);

///////////////////////////////////////////////////////////////////////////
//
// timer equations -
//

// CR_horizontal equations : "horizontal equals XXX" 
assign p1_h_eq_ht_m1=(p2_charctr_h== cr00);// charctr_h reached (h_total-1)
assign p2_h_eq_hbe= (p2_charctr_h == cr03);// charctr_h reached h_blank end
assign p2_h_eq_hre= (p2_charctr_h == cr05);// charctr_h reached h_resync end
assign p2_h_eq_hrs= (p2_charctr_h == cr04);// charctr_h reached h_resync start
assign p2_h_eq_hbs= (p2_charctr_h == cr02);// charctr_h reached h_blank start
assign p2_h_eq_hde= (p2_charctr_h == cr01);// charctr_h reached h_display end

// CR_vertical equations : "scanline equals YYY" 
assign p2_v_eq_vt_m1= (p3_linectr_v==cr06);// linectr_v reached (v_total-1)
assign p3_v_eq_vbe= (p3_linectr_v == cr16);// linectr_v reached vert_blank end
assign p3_v_eq_vre= (p3_linectr_v == cr11);// linectr_v reached v_resync end
assign p3_v_eq_vrs= (p3_linectr_v == cr10);// linectr_v reached v_resync start
assign p3_v_eq_vbs= (p3_linectr_v == cr15);// linectr_v reached v_blank start
assign p3_v_eq_vde= (p3_linectr_v == cr12);// linectr_v reached v_display end

  // pipeline registers
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    begin
      p2_h_eq_ht_m1 <= 1'b0;
      
      p3_h_eq_hbe<= 1'b0;
      p3_h_eq_hre<= 1'b0;
      p3_h_eq_hrs<= 1'b0;
      p3_h_eq_hbs<= 1'b0;
      p3_h_eq_hde<= 1'b0;
      p3_v_eq_vt_m1 <= 1'b0;

      p4_v_eq_vbe<= 1'b0;
      p4_v_eq_vre<= 1'b0;
      p4_v_eq_vrs<= 1'b0;
      p4_v_eq_vbs<= 1'b0;
      p4_v_eq_vde<= 1'b0;
    end
  else
    begin
      p2_h_eq_ht_m1 <= p1_h_eq_ht_m1;
      p3_h_eq_hbe<= p2_h_eq_hbe;
      p3_h_eq_hre<= p2_h_eq_hre;
      p3_h_eq_hrs<= p2_h_eq_hrs;
      p3_h_eq_hbs<= p2_h_eq_hbs;
      p3_h_eq_hde<= p2_h_eq_hde;
      p3_v_eq_vt_m1 <= p2_v_eq_vt_m1;
      
      p4_v_eq_vbe<= p3_v_eq_vbe;
      p4_v_eq_vre<= p3_v_eq_vre;
      p4_v_eq_vrs<= p3_v_eq_vrs;
      p4_v_eq_vbs<= p3_v_eq_vbs;
      p4_v_eq_vde<= p3_v_eq_vde;
    end

///////////////////////////////////////////////////////////////////////////
//
//  THE PIXEL/LINE COUNTERS
//
// The horizontal character counter (charctr_h) and vertical line counter
// (linectr_v) provide the basic monitor timing information.
//
// chartim_h is a sub-character timer (horizontal pixel)
//   The timer pulses once per character cell (character clock period.)
//   The timer's value is the X-position of the scan_beam within the 
//   character cell.
//
// charctr_h is the base (horizontal) character counter
//   charctr_h counts the # "character clocks" since the previous HSYNC.
//   charctr_h is the X-coordinate of the scan-beam, 
//   in units of character cells (text grid.)
//   The monitor signal timing (HSYNC|HSYNC) are derived from charctr_h.
//
// pixctr_h is an absolute horizontal pixclk counter
//   pixctr_h is a cumulative (horizontal) counter.
//   pixctr_h is the absolute X-coordinate of the scan-beam, in units of
//   pixclks.
//
// linectr_v is the scanline counter
//   linectr_v counts the number of HSYNCs that have elapsed since the
//   top-of-display (VRETRACE.)
//   linectr_v is the CRT-beam's (Y-coordinate) "scanline" position.
//   The monitor-control vo_vsync is derived from linectr_v.
//
// chartim_v is a sub-character timer (scanline)
//   chartim_v counts the # of scanlines within the character cell.  The
//   timer pulses once per character cell (vertical.)
//   It is the Y-position of the scan-beam within the character cell.
//
// charctr_v is the base (vertical) character counter.
//   charctr_v counts the # character cells, in the vertical (Y) direction,
//   since the previous top-of-display (VRETRACE.)
//
//    pixctr_h & chartim_v form the basic text-character cell size.
//    The actual text-character grid is defined by charctr_h X charctr_v
//
//////////////////////////////////////////////////////////////////////////////

    //                      _   _   _   _   _   _   _   _   _   _
    // pixclk             _|0|_|1|_|2|_|3|_|4|_|5|_|6|_|7|_|0|_|1|_...
    //                      _                               _
    // p1_chartim_h_reset _| |_____________________________| |_______________

  // chartim_h : horizontal "character-clock" signal (pixel accurate timer)
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    p1_chartim_h <= 0;
  else if ( p1_chartim_h_reset )  // reset after PIXELS_PER_CHAR pixclks
    p1_chartim_h <= 0;
  else if ( p1_chartim_h_inc )    // every pixel
    p1_chartim_h <= p1_chartim_h + 1;
  assign p1_chartim_h_inc   = 1'b1;   // period = (PIXELS_PER_CHAR+1) pixclks
  assign p1_chartim_h_reset = p1_chartim_h_inc && 
                              ( p1_chartim_h == PIXELS_PER_CHAR );
       
//////////////////////////////////////////////////////////////////////////////
//
// transfer P1 regs -> P2 regs
//

  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    begin
     p2_chartim_h_inc   <= 1'b0;
     p2_chartim_h_reset <= 1'b0;
     p2_chartim_h       <= 0;
    end
  else
    begin
     p2_chartim_h_inc   <= p1_chartim_h_inc;
     p2_chartim_h_reset <= p1_chartim_h_reset;
     p2_chartim_h       <= p1_chartim_h;
    end
  
  // pixctr_h (PIXELS_PER_CHAR MUST BE == 7!)
  assign p2_pixctr_h[9:3] = p2_charctr_h[6:0];
  assign p2_pixctr_h[2:0] = p2_chartim_h[2:0];

  // charctr_h : text-grid co-ordinate 'X' (horizontal character counter)
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    p2_charctr_h <= 0;
  else if ( p2_charctr_h_reset ) // reset at HORIZONTAL_TOTAL
    p2_charctr_h <= 0;
  else if ( p2_charctr_h_inc )   // increment on character-cell 
    p2_charctr_h <= p2_charctr_h + 1;
  assign p2_charctr_h_inc = (p2_chartim_h_reset);  // increment counter 
  assign p2_charctr_h_reset=(p2_charctr_h_inc && p2_h_eq_ht_m1); // reset ctr

    //                      _   _   _   _   _   _   _   _   _   _
    // pixclk             _|0|_|1|_|2|_|3|_|4|_|5|_|6|_|7|_|0|_|1|_...
    //              _______ _______________________________ _________________
    // p2_charctr_h _______|__n+1__________________________|__n+2____________

  // equivalent C-psuedocode
  //
  //  /* start of 1 scanline */
  //  for ( charctr_h = 0; charctr_h < ht; ++charctr_h )
  //  {
  //    for ( chartim_h = 0; chartim_h < 8; ++chartim_h )
  //    {};
  //    /* after the inner-loop, charctr has advanced by 1. */
  //  }
  //  /* after the outer-loop, charctr_h has advanced by 1. */
  //  /* end of 1 scanline period */

//////////////////////////////////////////////////////////////////////////////
//
// transfer P2 -> P3 regs
//
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    begin
      p3_charctr_h_inc   <= 1'b0;
      p3_charctr_h_reset <= 1'b0;
      p3_charctr_h   <= 0;
      p3_pixctr_h    <= 0;
    end
  else
    begin
      p3_charctr_h_inc   <= p2_charctr_h_inc;
      p3_charctr_h_reset <= p2_charctr_h_reset;
      p3_charctr_h   <= p2_charctr_h;
      p3_pixctr_h    <= p2_pixctr_h;
    end

  //assign p3_linectr_v[3:0] = p3_chartim_v[3:0];
  //assign p3_linectr_v[9:4] = p3_charctr_v[5:0];

  // linectr_v : (vertical) scanline counter
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    p3_linectr_v <= 0;
  else if ( p3_linectr_v_reset )  // reset at VERTICAL_TOTAL
    p3_linectr_v <= 0;
  else if ( p3_linectr_v_inc )    // increment at start of new scanline
    p3_linectr_v <= p3_linectr_v + 1;
  assign p3_linectr_v_inc = (p3_charctr_h_reset);// increment counter 
  assign p3_linectr_v_reset=(p3_linectr_v_inc && p3_v_eq_vt_m1);// reset ctr

  // chartim_v : character cell, (Y)-position within cell (scanline accurate)
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    p3_chartim_v <= 0;
  else if ( p3_chartim_v_reset )  //  reset at character-cell boundary 
    p3_chartim_v <= 0;
  else if ( p3_chartim_v_inc )    // increment at start of new scanline
    p3_chartim_v <= p3_chartim_v + 1;
  assign p3_chartim_v_inc   = p3_linectr_v_inc;
  assign p3_chartim_v_reset = ( p3_chartim_v_inc && (p3_chartim_v == cr09) )
                              || p3_linectr_v_reset ;
//         (p3_chartim_v == `DFV_REG_CR09) ) || p3_linectr_v_reset ;

  // charctr_v : character row counter, (Y)-position within text-screen GRID
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    p3_charctr_v <= 0;
  else if ( p3_charctr_v_reset )  // reset at VERTICAL_TOTAL
    p3_charctr_v <= 0;
  else if ( p3_charctr_v_inc )    // increment at character-cell boundary
    p3_charctr_v <= p3_charctr_v + 1;
  assign p3_charctr_v_inc = p3_chartim_v_reset;    // increment counter
  assign p3_charctr_v_reset = p3_linectr_v_reset;  // reset counter

  // equivalent C-psuedocode
  //
  //  /* start of display cycle : beam is at top-left corner of CRT */
  //
  // :top_of_display
  //
  //  charctr_v = 0; /* scanline-total counter */
  //  for ( linectr_v = 0; linectr_v < vt; ++linectr_v )
  //  {
  //    for ( chartim_v = 0; chartim_v < cr09; ++chartim_v )
  //    {
  //       /* start of 1 scanline */
  //         /* insert the charctr_h code from above! */
  //       /* end of 1 scanline */
  //    }
  //    /* after the inner-loop, chartim has advanced by 1. */
  //
  //    ++charctr_v;  /* one character has passed */
  //
  //  } /* end for ( linectr_v ) */
  //
  //  /* after the outer-loop, we have completed 1 screen refresh! */
  //
  // goto :top_of_display;
  //

//////////////////////////////////////////////////////////////////////////////
//
// pixena_h/pixena_v : "pixel display enable"
//   Together, these two state-bits mark the 'active display area.' 
// The 'active display area' is the visible portion of the screen.  
//
// When either (or both) state-bits are low, pixel-display is disabled for
// that corresponding portion of the screen.  A 'blanking interval' is the
// time between two successive scanlines, when the display is disabled.
// Blanking intervals allow the CRT electron-beam opportunity to move from
// the end of the current scanline to the start of the next scanline.
//
//////////////////////////////////////////////////////////////////////////////

  // pixena_h : enable pixel scanout (horizontal)
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    p3_pixena_h <= 1'b0;  // horizontal active area
  else if ( p3_pixena_h_set  )   // activate at start of display
    p3_pixena_h <= 1'b1;
  else if ( p3_pixena_h_reset )  // disable at horizontal_display_end
    p3_pixena_h <= 1'b0;
  assign p3_pixena_h_set   = p3_charctr_h_reset; // enable at scanline start
  assign p3_pixena_h_reset = p3_charctr_h_inc && p3_h_eq_hde;

  // the raw HPULSE signal (active HIGH, not active low!)
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    p3_hpulse <= 1'b0;     // deactivate hsync pulse
  else if ( p3_hpulse_set ) // reached hretrace start? 
    p3_hpulse <= 1'b1;     // activate hsync pulse
  else if ( p3_hpulse_reset ) // reached hretrace end? 
    p3_hpulse <= 1'b0;     // deactivate hsync pulse
  assign p3_hpulse_set = p3_charctr_h_inc && p3_h_eq_hrs;//reach hretrace start?
  assign p3_hpulse_reset = p3_charctr_h_inc && p3_h_eq_hre;//reach hretrace end?

  // equivalent C-psuedocode : add to the earlier charctr_h code!
  // 
  //  /* start of 1 scanline */
  //  pixena_h = 1; /* start of scanline, entering active-display area */
  //
  //  for ( charctr_h = 0; charctr_h < ht; ++charctr_h )
  //  {
  //    for ( chartim_h = 0; chartim_h < 8; ++chartim_h )
  //    {};
  //    /* after the inner-loop, charctr has advanced by 1. */
  //
  //    if ( charctr_h == hde ) /* reach horizontal_display_end? */
  //      pixena_h = 0;  /* yes, outside of active-display area */
  //  }
  //  /* after the outer-loop, charctr_h has advanced by 1. */
  //
  //  /* end of 1 scanline period */

//////////////////////////////////////////////////////////////////////////////
//
// transfer P3 -> P4 regs
//
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    begin
      p4_charctr_h_inc   <= 1'b0;
      p4_charctr_h_reset <= 1'b0;
      p4_pixena_h   <= 1'b0;
      p4_hpulse <= 1'b0;
      
      p4_linectr_v_inc   <= 1'b0;
      p4_linectr_v_reset <= 1'b0;
      p4_chartim_v_reset <= 1'b0;
      p4_chartim_v_inc <= 1'b0;
    end
  else
    begin
      p4_charctr_h_inc   <= p3_charctr_h_inc;
      p4_charctr_h_reset <= p3_charctr_h_reset;
      p4_pixena_h   <= p3_pixena_h;
      p4_hpulse <= p3_hpulse;

      p4_linectr_v_inc   <= p3_linectr_v_inc;
      p4_linectr_v_reset <= p3_linectr_v_reset;
      p4_chartim_v_reset <= p3_chartim_v_reset;
      p4_chartim_v_inc <= p3_chartim_v_inc;
    end

  // pixena_v : enable pixel scanout (vertical)
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    p4_pixena_v <= 1'b0;  
  else if ( p4_pixena_v_set )    // activate at top of display
    p4_pixena_v <= 1'b1;
  else if ( p4_pixena_v_reset )  // reset at VERTICAL_DISPLAY_END
    p4_pixena_v <= 1'b0;
  assign p4_pixena_v_set   = p4_linectr_v_reset;
  assign p4_pixena_v_reset = p4_linectr_v_inc && p4_v_eq_vde;

  // the raw VPULSE signal (active HIGH, not active low)
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    p4_vpulse <= 1'b0;     // deactivate vsync pulse
//  else if ( v_eq_vrs ) // reached vretrace start?
  else if ( p4_vpulse_set ) // reached vretrace start? 
    p4_vpulse <= 1'b1;     // activate vsync pulse
//  else if ( v_eq_vre ) // reached vretrace end?
  else if ( p4_vpulse_reset ) // reached vretrace end? 
    p4_vpulse <= 1'b0;     // deactivate vsync pulse
  assign p4_vpulse_set = p4_linectr_v_inc && 
                       p4_v_eq_vrs; // reached vretrace start? 
  assign p4_vpulse_reset = p4_linectr_v_inc &&
                         p4_v_eq_vre; // reached vretrace end? 
  assign p4_scanena = p4_pixena_h && p4_pixena_v; // form 'active display'

  // equivalent C-psuedocode: add to the earlier charctr_v code!
  //
  //  /* start of display cycle : beam is at top-left corner of CRT */
  //
  // :top_of_display
  //
  //  charctr_v = 0; /* scanline-total counter */
  //  pixena_v = 1;  /* top_of_display, entering active-display area! */
  //
  //  for ( linectr_v = 0; linectr_v < vt; )
  //  {
  //    for ( chartim_v = 0; chartim_v < cr09; ++chartim_v )
  //    {
  //       /* start of 1 scanline */
  //         /* insert the charctr_h code from above! */
  //       /* end of 1 scanline */
  //       ++linectr_v;
  //
  //       if ( linectr_v == vde ) /* reach vertical_display_end? */
  //         pixena_v = 0;  /* yes, outside of active-display area */
  //    }
  //    ++charctr_v;  /* one character has passed */
  //
  //    /* after the inner-loop, chartim has advanced by 1. */
  //  } /* end for ( linectr_v ) */
  //
  //  /* after the outer-loop, we have completed 1 screen refresh! */
  //
  // goto :top_of_display;
  //

//
//  end of PIXEL/LINE counters
//
/////////////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////////////
//
// Framebuffer scan
//
// This portion of the CRTC-module generates memory address information 
// to access the framebuffer.  The retrieved data is a single byte of
// (ASCII) 'character code.'  The character-code is then used as a pointer
// index into the font-bitmap memory.
//
//    The address-generator contains 2 stages: fbbase and fbaddr.
// fbbase is the "scanline-base-address."  It points to the framebuffer
// address of the leftmost character of the current scanline.
// fbaddr is the working address pointer.  It points to the fraembuffer
// address of the character under the video-beam.
//
//    The 'fbaddr' (framebuffer current address) is 1T delayed relative to 
// the p4_XXX base pixel/char counters.  The 'fbbase' (scanline base 
// address) operates in the p4_stage.  fbaddr operates 1-pipeline stage 
// later (f0_stb.)


// fbs_fbbase = 'scanline base address', updated every scanline
// at vsync, fbs_fbbase is reset to start of screen address (top of screen)
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    begin
      fbs_fbbase[15:8] <= 0;  // high address
      fbs_fbbase[7:0] <= 0;   // low address
    end
  else if ( fbs_fbbase_reset )  // reset fbbase at vsync (to top of screen)
    begin
      fbs_fbbase[15:8] <= cr0c;  // high address field
      fbs_fbbase[7:0] <= cr0d;   // low address field
    end
  else if ( fbs_fbbase_inc )  // inc framebuffer address by one scanline
    begin
      fbs_fbbase <= fbs_fbbase_next_1e; 
    end
  assign fbs_fbbase_inc = p4_chartim_v_reset && p4_pixena_v;
  assign fbs_fbbase_reset = p4_linectr_v_reset; 
  

  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    fbs_fbbase_next_1e <= 0;
  else
    fbs_fbbase_next_1e <= fbs_fbbase_next; 
  assign fbs_fbbase_next = fbs_fbbase + cr13; // add +pitch (1-scanline)

  //  pseudocode-C
  //
  // :top_of_display; /* beam is in upper-left corner, entering active! */
  //  fbs_fbbase = make_abs_address(cr0c, cr0d); /* initialize address */
  //  linectr_v = 0;
  //  
  //  while ( linectr_v < vde )
  //  {
  //    fbs_fbbase_next = fbs_fbbase + cr13; /* NEXT scanline address */
  //
  //    for ( chartim_v = 0; chartim_v < cr09; ++chartim_v )
  //    {
  //       /* start of 1 scanline */
  //         /* insert the charctr_h code from above! */
  //       /* end of 1 scanline */
  //       ++linectr_v;
  //
  //       if ( linectr_v == vde ) /* reach vertical_display_end? */
  //         break;  /* yes, terminate this inner-loop early! */
  //    }
  //    ++charctr_v;  /* one character has passed */
  //
  //    if ( linectr_v == vde ) /* reach vertical_display_end? */
  //      break;  /* yes, terminate this outer-loop early! */
  //
  //    /* after the inner-loop, chartim has advanced by 1. */
  //    fbs_fbbase = fbs_fbbase_next; /* load next scanline address */
  //  } /* end for ( linectr_v ) */
  //
  //  /* after the outer-loop, we have completed 1 screen refresh! */
  //
  // goto :top_of_display;

/////////////////////////////////////////////////////////////////////////////
//
// control signal generation for fbaddr, fbs_scanena
//

  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    begin
      fbs_fbaddr_charinc<= 1'b0; 
      fbs_fbaddr_reload <= 1'b0;
      fbs_scanena <= 1'b0;
    end
  else
    begin
      fbs_fbaddr_charinc<= fbs_fbaddr_charinc_1e;// increment 1 character
      fbs_fbaddr_reload <= fbs_fbaddr_reload_1e;// reload at start of scanline
      fbs_scanena <= p4_scanena;
    end
  assign fbs_fbaddr_charinc_1e = (p4_charctr_h_inc && p4_pixena_h);
  assign fbs_fbaddr_reload_1e  = p4_charctr_h_reset; 

  // fbs_curaddr is updated every character 
  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    fbs_fbaddr <= 0;
  else if ( fbs_fbaddr_reload )  // load curraddr from fbs_fbbase
    fbs_fbaddr <= fbs_fbbase; // add offset pitch
  else if ( fbs_fbaddr_charinc ) // increment 1 character
    fbs_fbaddr <= fbs_fbaddr + 1; // updated on F1
  
  //  /* this loop runs in PARALLEL (concurrent execution) with above loop */
  //
  // :top_of_display; /* beam is in upper-left corner, entering active! */
  //  linectr_v = 0;
  //
  //  while ( linectr_v < vde  )
  //  {
  //    /* start of a new scanline */
  //    /* remember, fbs_fbbase is updated at start of each scanline! */
  //    fbs_fbaddr = fbs_fbbase; /* START (leftmost char) of scanline */
  //
  //    /* this inner loop reads a row of characters from RAM. */
  //
  //    for( charctr_h=0; charctr_h < hde; ++charctr_h )
  //    {
  //      ++fbs_fbaddr; /* got fbs_fbaddr_charinc */
  //      text_data = perform_memory_access( fbs_fbaddr ); /* get char */
  //    }
  //
  //    /* reached end of scanline, fbs_fbaddr_reload */
  //  }
  

// The pipeline-stages F0-F7 form one complete horizontal character period
// For example, |F0 F1 F2 F3 F4 F5 F6 F7|F0 F1 F2 F3 F4 F5 F6 F7|
//              |CHARACTER_CELL N       |  CHARACTER_CELL N+1   |
//
// Stage P1 P2 P3 P4 P5 P6 P7 P8|P1 P2 P3...
//                   F0 F1 F2 F3 F4
//                   ^^ gated on pixena_h -
// Stages labeled 'F0' and later are gated against pixena_h.  In other
// words, f0_stb is inactive (low) for the entire horizontal retrace period
// of every scanline!  This causes a few anomalous behaviors outside the
// active display area.  But we don't worry about it, because the 
// CRT-scan inactive during the retrace-period.  

// f2_text_data = fetch caracter code
always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    begin
      f1_stb <= 1'b0;
      f2_stb <= 1'b0;
      f3_stb <= 1'b0;
      f2_text_data <= 0;
    end
  else
    begin
      f3_stb <= f2_stb;  // f3_stb is high when f2_text_data has new data
      f2_stb <= f1_stb;
      f1_stb <= f0_stb;
      if ( f1_stb )
//        f2_text_data <= {p3_linectr_v[7:4],p2_charctr_h[3:0]}; 
        f2_text_data <= i_sramd; // get character-code from framebuffer
    end
assign f0_stb =fbs_fbaddr_charinc;  // misaligned (line ends at very left)

    //           _   _   _   _   _   _   _   _   _   _
    // pixclk  _|0|_|1|_|2|_|3|_|4|_|5|_|6|_|7|_|0|_|1|_...
    //           _                               _
    // f3_stb  _| |_____________________________| |_______________
    //           _______________________________ ____________
    // text_data| character code - (X,Y)        |  (X+1,Y)
//
// end of Framebuffer scan
//
/////////////////////////////////////////////////////////////////////////////


  always @ ( posedge pixclk or negedge rstn )
  if ( !rstn )
    begin
      o_srama <= 0;
      o_sram_reqn <= 1'b0;
    end
  else
    begin
      o_srama <= o_srama_next; // 1T delay relative to o_srama_next
      o_sram_reqn <= !(f0_stb || f1_stb); // becomes f1_stb + f2_stb
    end
  assign o_srama_next = o_srama_sel ? fbs_fbaddr : font_addr; // next-state
  assign o_srama_sel = f0_stb;
// 50MHz PIXCLK (Xess XS-40 board, asynchronous SRAM 15ns)
//   each memory access requires 2pixclks
//assign o_srama_sel = f0_stb || f1_stb;
// stages f1-f2 for f2_text_data (character)
// stages f3-f4 for f3_font_data (font bitmap)

// 25MHz PIXCLK (Xess XS-40 board, asynchronous SRAM 15ns)
//   each memory access requires 1pixclk
//assign o_srama_sel = f0_stb;
// SRAM access furing F1 to fetch character (text_data or character_code)
// SRAM access during f2 to fetch f3_font_data (character bitmap)


/////////////////////////////////////////////////////////////////////////////
//
// Font-Map Address generator
//
// Using the retrieved character-code, the fontmap address calculator
// schedules one more memory access.  The fetched font data is a 'slice'
// of the character-code bitmap.  The bitmap is a binary-coded 2D-image.
// The output of the address-generator is an absolute address $0000-$FFF
// (0-4095decimal)  The code below assumes the font-map is located at
// address location $00000-$00FFF.  

  // Address-generator operation :
  //
  // 1) Calculate Z = (CHAR_CODE * CHAR_HEIGHT);
  // 2) Calculate final_addr = Z + Y_position;   // p4_chartim_v
  
  assign font_inx = f2_text_data; 
  assign font_iny = cr09+1;  // CR09 = "char height minus 1"
  assign font_start = f2_stb;
  assign font_inso = {3'h0, p3_chartim_v}; // scan-line offset, not used


// This multiplier calculates the address_offset into the font bitmap.
// For fixed VGA-compatible font-size (char_height=16), a multiplier isn't
// needed, a simple bit-shift does the job!

/*  expected latency is 6 pixclks (almost one char_clock) */
  mult8x5u font_addrgen( 
  // inputs
//    .i_x(font_inx), .i_y(font_iny), .init_val(font_inso), .clk(pixclk), 
    .i_x(font_inx), .i_y(font_iny), .clk(pixclk), .rstn(rstn), 
    .i_start(font_start), 
  // outputs
    .o_z(font_outz), .o_done(font_done), .o_busy(font_busy)
    );
    

  // why "&& p4_pixena_h" ? p4 stage comes before f0, the and-operation
  // surpresses the 1st f0_stb on the scanline (right after 
  // p4_charinc_reset)  
  //           _   _   _   _   _   _   _   _   _   _   _
  // pixclk  _| |_|0|_|1|_|2|_|3|_|4|_|5|_|6|_|7|_|0|_|1|_...
  // fbs_addr  _ <-----start of new scanline
  // reload  _| |___________________________________________________
  //               _                               _
  // f0_stb  _____| |_____________________________| |_______________
  //                       _                               _
  // f2_stb  _____________| |_____________________________| |_______
  //               .                                _
  // ***     _____________________________________| |_______________
  //              ^^^ need to kill the 1st pulse!
//  assign font_done = f0_stb && (!fbs_fbaddr_reload);  // f2_stb -> f0_stb = 6 pixclks
//  assign font_outz = { 1'b0, font_inx, 4'h0 };
//  assign font_busy = 1'b0;

  always @ ( posedge pixclk or negedge rstn )
    if ( !rstn )
      f1_font_addr <= 0;
    else if ( font_done ) // font_done = f0_stb
      f1_font_addr <= font_outz + p3_chartim_v; // 2) final_addr
  assign font_addr = {sr03, f1_font_addr[12:0] }; // f3_stb

  // for 16-tall characters only!
  //assign font_addr = {4'h0, f2_text_data, p4_chartim_v[3:0] }; 

  // although p3_chartim_v and font_done are from different pipeline
  // stages, p3_chartim_v only changes once per scanline.  It changes at
  // the very end of the horizontal scanline period, where it won't
  // adversely affect the above calculation.

  // font_addr : The 13-bit address space covers the entire font-map :
  //   0x0000-0x1FFF
  //
  // font_addr is ready during cycle 'f1.'  Due to the address-mux-reg
  // in this module, font_addr doesn't reach the sram_address lines, 
  // until 1T later (f2_stb.)

  // note that f2_stb actually loads 1char before the start of a new
  // scanline.  That's because the multiplier-latency introduces a skew of
  // +1 character period.  (so the correct loading-point for f3_font_data 
  // is an 8pixclk delayed version of f2_stb.  Not to worry, the ds_scanena
  // blacks out the bogus character from the final output.)
  
  always @ ( posedge pixclk or negedge rstn )
    if ( !rstn )
      f3_font_data <= 0;
    else if ( f2_stb ) // should be font_done_2d
      f3_font_data <= i_sramd;
    else
      f3_font_data <= f3_font_data << 1; // shift data out every pixclk

//
// end of Font-Map Address generator
// 
/////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
// RGB pixel-value scanout
//

  // Convert (monochrome) font bitmap into R, G, B intensity values. 

`define WHITE 6'b11_11_11
`define GREY2 6'b10_10_10
`define GREY1 6'b01_01_01
`define BLACK 6'b00_00_00

  // RGB pixel output
  always @ ( posedge pixclk or negedge rstn )
    if ( !rstn )
      begin
        vo_pixeldata <= `BLACK; 
      end
    else
      begin
        if ( ds_scanena[2] )  // was 2, with font_done = f0_stb 
          vo_pixeldata <= { vo_pixeldataR, vo_pixeldataG, vo_pixeldataB };
        else 
          vo_pixeldata <= `BLACK;  // outside of 'active area', force to BLACK
      end

  // convert single font_data pixel-bit into R,G,B color-value
  assign vo_pixeldataR = f3_font_data[ 7 ] ? 2'b11 : 2'b00;
  assign vo_pixeldataG = f3_font_data[ 7 ] ? 2'b11 : 2'b00;
  assign vo_pixeldataB = f3_font_data[ 7 ] ? 2'b11 : 2'b00;

//
// end of Framebuffer scan
//
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
//
// deskew delay line
//

  // Recall that the cumulative latency in pixclks, measured from "p1_stage"
  // to final-output (vo_pixeldata) equals 10-15 pixclks. 
  // The hsync/vsync lines are generated during stage_p4, while the RGB 
  // intensity values the generated at the tail end (10-15 pixclk.)  
  // The deskewer coarsely realigns the HSYNC/VSYNC pulses with the
  // the R,G,B signal lines.  
  //
  // Persisting skew manifests itself as a horizontal display offset,
  // which is visible as a left/right pan.  
  
  always @ ( posedge pixclk or negedge rstn )
    if ( !rstn )
      begin
        ds_hsync <= 4'b1111;
        ds_vsync <= 4'b1111;
      end
    else if ( p4_charctr_h_inc )
      begin
        ds_hsync[0] <= ~p4_hpulse;      // hsync_n, convert to active-low
        ds_vsync[0] <= ~p4_vpulse;      // vsync_n, convert to active-low
        ds_hsync[3:1] <= ds_hsync[2:0];
        ds_vsync[3:1] <= ds_vsync[2:0];
      end
      

  // Note that ds_scanena 
  always @ ( posedge pixclk or negedge rstn )
    if ( !rstn )
      ds_scanena <= 0;
    else if ( f2_stb ) // sync'd with f3_font_data loader
      ds_scanena <= { ds_scanena[2:1], fbs_scanena };
    
//////////////////////////////////////////////////////////////////////////////
//
// output signal generation
//
// vo_hsync/vo_vsync are 2T delayed relative to pixctr
// both vo_hsync/vo_vsync are active low

  assign o_vga_hsync_n = ds_hsync[1];  //1 = just right
  assign o_vga_vsync_n = ds_vsync[1];  //2 too far to left
  assign o_vga_pixeldata = vo_pixeldata;

endmodule
